from __future__ import annotations

PERSIST_UTILITY = 0x1
PERSIST_EXP_UTILITY = 0x2
PERSIST_PROBABILITY = 0x4
PERSIST_LOGLIKE_CASEWISE = 0x8

PERSIST_D_UTILITY = 0x10
PERSIST_COND_LOG_PROB = 0x20
PERSIST_D_PROBABILITY = 0x40
PERSIST_D_LOGLIKE_CASEWISE = 0x80

PERSIST_BHHH = 0x100
PERSIST_PARTS = 0x200
PERSIST_QUANTITY = 0x400

PERSIST_ALL = 0xFFFF
